unit UnitConvert;

interface

uses KOL;

procedure ConvertFile( const fname, SecondUsesList: String );
function BuildUses2( const module: String ): String;

var
    Step: Integer; // = 1:   ,   -     
                   // = 2:  ,  ,     
                   //       uses #2

implementation

//  {$DEFINE TEST}

uses Windows {$IFDEF TEST},
  tif_dirread, tif_dir, tif_aux, tif_getimage, tif_error, tif_zip, tif_luv,
  tif_compress, tif_codec, tif_close, tif_version, tif_thunder, tif_swab,
  tif_jpeg, tif_ojpeg,
  tif_read, tif_write, tif_strip, tif_tile, tif_lzw, tif_next,
  tif_packbits, tif_pixarlog, tif_warning, tif_flush, tif_fax3,
  tif_dirwrite, tif_dumpmode, tif_predict
  {$ENDIF};

var EL, BL: PStrList;
    ML: PStrListEx; //  
    is_data_segment: Boolean;

function FirstWord( const s: String ): String;
var tmp: String;
begin
  tmp := trim( s );
  Result := Parse( tmp, ' '#9 );
end;

function SecondWord( const s: String ): String;
var tmp: String;
begin
  tmp := s;
  Parse( tmp, ' '#9 );
  tmp := Trim( tmp );
  Result := Parse( tmp, ' '#9';:,' );
end;

function PartBeforeSemicolon( var s: String ): String;
var i: Integer;
    in_string: Boolean;
begin
  in_string := FALSE;
  for i := 1 to Length( s ) do
  begin
    CASE s[i] OF
    '''': in_string := not in_string;
    ';' : if not in_string then
          begin
            Result := Copy( s, 1, i-1 );
            s := CopyEnd( s, i );
            Exit;
          end;
    END;
  end;
  Result := s;
  s := '';
end;

function ReplaceSymbols( const s: String ): String;
var i, j, k: Integer;
    w: String;
    found_bad_chars: Boolean;
    is_local_label: Boolean;
begin
  Result := s;
  i := 1;
  while i <= Length( Result ) do
  begin
    if Result[ i ] in [ ' ', #9, ',', ':', ';', '-', '+',
                        '[', ']', '(', ')', '/', '*' ] then inc( i )
    else
    begin //  
      found_bad_chars := FALSE;
      j := i;
      while j <= Length( Result ) do
      begin
         if Result[ j ] in [ ' ', #9, ',', ':', ';', '-', '+',
                             '[', ']', '(', ')', '/', '*' ] then
         begin //   
           break;
         end
           else
         if Result[ j ] in [ '@', '$', '?' ] then
           found_bad_chars := TRUE;
         inc( j );
      end;
      is_local_label := FALSE;
      if (Result[ i ] = '$') and (Result[ i+1 ] = 'L') and
        (Result[ i+2 ] in [ '0'..'9' ]) or
        (Result[ i ] = '$') and (Result[ j-1 ] in [ '0'..'9' ]) then
      begin //  :
        is_local_label := TRUE;
        if BL.Count > 0 then
        begin
          w := Copy( Result, i, j-i );
          if BL.IndexOf( w ) >= 0 then
          begin
            is_local_label := FALSE;
            Result := Copy( Result, 1, i-1 ) + '_' + CopyEnd( Result, i );
            inc( j );
          end;
        end;
      end;
      if is_local_label then
      begin
        //Result[ i ] := '@';
        for k := i to j do
          if Result[ k ] = '$' then Result[ k ] := '@';
      end
       else
      begin
        if found_bad_chars then
        begin
          w := Copy( Result, i, j-i );
          k := 1;
          while k <= Length( w ) do
          begin
            CASE w[ k ] OF
            '$': w := Copy( w, 1, k-1 ) + 'd_' + CopyEnd( w, k+1 );
            '@': w := Copy( w, 1, k-1 ) + 'a_' + CopyEnd( w, k+1 );
            '?': w := Copy( w, 1, k-1 ) + 'q' + CopyEnd( w, k+1 );
            END;
            inc( k );
          end;
          if Length( w ) > 64 then
           w := Copy( w, 1, 64 );
          Result := Copy( Result, 1, i-1 ) + w + CopyEnd( Result, j );
          j := i + Length( w );
        end;
        {if Result[ i ] = '_' then //    
        begin
          w := Copy( Result, i, j-i );
          if EL.IndexOf( w ) >= 0 then
          begin
            Delete( Result, i, 1 );
            dec( j );
          end;
        end;}
      end;
      i := j;
    end;
  end;
end;

function AsmOperator( const s: String ): String;
var tmp, w1: String;
begin
  tmp := s;
  if tmp = '' then
  begin
    Result := ''; Exit;
  end;
  w1 := Parse( tmp, ' '#9 );
  if StrIn( w1, [ 'DB', 'DW', 'DD', 'DQ' ] ) then
  begin
    Result := w1;
    Exit;
  end;
  Result := Trim( Parse( tmp, ' '#9 ) );
end;

function RemoveShort( const s: String ): String;
var i: Integer;
begin
  Result := s;
  i := pos( 'SHORT', UpperCase( s ) );
  if i > 1 then
  begin
    if (s[ i+5 ] in [ ' ', #9 ]) and
       (s[ i-1 ] in [ ' ', #9 ]) then
    begin
      Result := Copy( s, 1, i-1 ) + CopyEnd( s, i+6 );
    end;
  end;
end;

function ReplaceOrg( const s: String ): String;
var i, n: Integer;
    least: String;
begin
  Result := s;
  i := pos( 'ORG', s );
  if i <= 0 then i := pos( 'ORG', UpperCase( s ) );
  if i <= 0 then Exit;
  least := Trim( CopyEnd( s, i+3 ) );
  if least = '' then Exit;
  if (least[ 1 ] = '$') and (least[ 2 ] = '+') then
  begin
    least := CopyEnd( least, 3 );
    n := Str2Int( least );
    CASE n OF
    1: Result := Copy( s, 1, i-1 ) + 'DB 0';
    2: Result := Copy( s, 1, i-1 ) + 'DW 0';
    4: Result := Copy( s, 1, i-1 ) + 'DD 0';
    else
    begin
      Result := Copy( s, 1, i-1 ) + 'DB ' + Int2Str( n ) + '(0)';
    end;
    END;
  end;
end;

function RemoveTrailingChar( const s: String; c: Char ): String;
begin
  Result := s;
  if (s <> '') and (s[ Length( s ) ] = c) then
    Delete( Result, Length( s ), 1 );
end;

function ReplaceRepeat( const s: String; before: Integer; const data_directive: String ): String;
var i, n: Integer;
    is_hex: Boolean;
    s1, s2, s3, s4, portion: String;
    SL: PStrList;
begin
  Result := s;
  i := before-1;
  if i <= 0 then Exit;
  is_hex := FALSE;
  while (i > 1) and (s[ i ] in [ ' ', #9 ]) do dec( i );
  if s[i] in [ 'h', 'H' ] then is_hex := TRUE;
  while (i > 0) and not(s[i] in [ ' ', #9 ]) do dec( i );
  if (i = 0) or (s[ i ] in [ ' ', #9 ]) then inc( i );
  if is_hex then n := Hex2Int( CopyEnd( s, i ) )
            else n := Str2Int( CopyEnd( s, i ) );
  s1 := Copy( s, 1, i-1 );
  s4 := Trim( CopyEnd( s, before ) );
  s3 := PartBeforeSemicolon( s4 );
  if Length( s3 ) < Length( s4 ) then
    s4 := '// ' + CopyEnd( s4, Length( s3 ) + 1 )
  else
    s4 := '';
  s2 := Trim( Parse( s3, ',' ) );
  SL := NewStrList;
  TRY
    portion := s1;
    for i := 1 to n do
    begin
      portion := portion + s2 + ',';
      if Length( portion ) > 64 then
      begin
        SL.Add( RemoveTrailingChar( portion, ',' ) );
        portion := ' ' + data_directive + ' ';
      end;
    end;
    portion := TrimRight( portion + s3 );
    portion := RemoveTrailingChar( portion, ',' );
    if portion <> '' then
      SL.Add( portion );
    Result := TrimRight( SL.Text );
    Result := RemoveTrailingChar( Result, ',' ) + s4;
  FINALLY
    SL.Free;
  END;
end;

function ReplaceDQR( const s, data_directive: String ): String;
var i, j: Integer;
    num, least: String;
begin
  Result := s;
  i := pos( data_directive, s );
  if i <= 0 then i := pos( data_directive, UpperCase( s ) );
  if i <= 0 then Exit;
  num := Trim( CopyEnd( s, i+3 ) );
  j := pos( ';', num );
  if j <= 0 then
  begin
    least := '';
  end
    else
  begin
    least := ''#9 + CopyEnd( num, j );
    num := Trim( Copy( num, 1, j-1 ) );
  end;
  if (num <> '') and (num[ Length( num ) ] = 'r') then
    num := Copy( num, 1, Length( num )-1 ) + 'H';
  Result := Copy( s, 1, i+2 ) + num + least;

  // remove DUP
  i := pos( 'DUP', Result );
  if (i > 1) and (Result[ i-1 ] in [ ' ', #9 ]) and
     (Result[ i+3 ] in [ ' ', #9, '(' ]) then
  begin
     Delete( Result, i, 3 );
     j := pos( '(?)', Result );
     if j > 0 then
       Result[ j+1 ] := '0'
     else
     asm
       int 3
     end;
     Result := ReplaceRepeat( Result, i, data_directive );
  end;
end;

function RemoveFlat( const s: String ): String;
var i: Integer;
begin
  Result := s;
  REPEAT
    i := pos( 'FLAT:', Result );
    if (i > 1) then
      if (s[ i-1 ] in [ ' ', #9 ]) then
        Delete( Result, i, 5 )
      else
        break;
  UNTIL i <= 0;
end;

function OperandNumber( const s: String ): Boolean;
var t: String;
    i: Integer;
begin
  t := s;
  t := Trim( Parse( t, ';' ) );
  Result := FALSE;
  if t = '' then Exit;
  for i := 1 to Length( t ) do
    if not ( t[i] in [ '0'..'9' ] ) then Exit;
  Result := TRUE;
end;

function SwapTestOperands( const s: String ): String;
var i: Integer;
    opds, L, r: String;
begin
  Result := s;
  i := pos( 'test', s );
  if i <= 0 then i := pos( 'TEST', UpperCase( s ) );
  if i <= 0 then Exit;
  opds := CopyEnd( s, i+4 );
  r := Trim( opds );
  L := Parse( r, ',' );
  if StrIn( L, [ 'eax', 'ebx', 'ecx', 'edx', 'ebp', 'esi', 'edi' ] ) and
     not OperandNumber( r ) then
  begin
    Parse( opds, ',' );
    opds := Parse( opds, ';' );
    Result := Copy( s, 1, i+4 ) + opds + ',' + L;
  end;
end;

function RepeatLongStr( const preambla, prefix: String; n: Integer; const r: String ): String;
var L: Integer;
    RL: PStrList;
    s: String;
begin
  L := Length( preambla );
  RL := NewStrList;
  TRY
    s := '';
    while n > 0 do
    begin
      if (Length( s ) + L > 64) and (s <> '') then
      begin
        RL.Add( RemoveTrailingChar( s, ',' ) );
        L := 0;
        s := prefix;
      end;
      s := s + r;
      Dec( n );
    end;
    if s <> '' then
      RL.Add( s );
    Result := TrimRight( RL.Text );
  FINALLY
    RL.Free;
  END;
end;

function ReplaceNPAD( const s: String ): String;
var i, n: Integer;
    t, r: String;
begin
  Result := s;
  i := pos( 'npad', s );
  if i <= 0 then i := pos( 'NPAD', UpperCase( s ) );
  if i <= 0 then Exit;
  t := Trim( CopyEnd( s, i+4 ) );
  n := Str2Int( t );
  CASE n OF
  1: r := 'nop';
  2: r := 'mov eax, eax';
  3: r := 'mov eax, eax'#13#10' nop';
  4: r := 'mov eax, eax'#13#10' mov eax, eax';
  5: r := 'mov eax, eax'#13#10' mov eax, eax'#13#10' nop';
  6: r := 'mov eax, eax'#13#10' mov eax, eax'#13#10' mov eax, eax';
  else if n < 128 then
       begin
         r := 'db 0ebh, ' + Int2Str( n - 2 ) + ', ';
         r := r + RepeatLongStr( r, ' DB ', n-2, '(90h),' )
       end
       else
       begin
         r := 'db 0e9h, ' + Int2Str( (n - 3) and 255 ) + ', ' +
                            Int2Str( (n - 3) shr 8 ) + ', ';
         r := r + RepeatLongStr( r, ' DB ', n-3, '(90h),' );
       end;
  END;
  Result := Copy( s, 1, i-1 ) + RemoveTrailingChar( r, ',' ) + ' // npad ' + t;
end;

function CommArrayDeclaration( const s: String ): String;
var t, a, b: String;
begin
  t := s;
  Parse( t, ' '#9 ); //  COMM
  t := trim( t );
  a := ReplaceSymbols( Trim( Parse( t, ':' ) ) ); //  
  Result := 'var ' + a + ': array[0..';
  b := Trim( Parse( t, ':' ) );
  if not StrEq( b, 'BYTE' ) then
  asm
    int 3 //    BYTE     
  end;
  t := Trim( t ); // t =  
  Result := Result + Int2Str( Hex2Int( t )-1 ) + '] of ' + b + ';';
end;

procedure CreateEmptyFile( const s: String );
var FL: PStrList;
begin
  FL := NewStrList;
  FL.SaveToFile( s );
  FL.Free;
end;

function ParamList( CL: PStrList ): String;
var i, j, k, off: Integer;
    s, param: String;
begin
  Result := '';
  if CL.Count = 0 then Exit;
  j := 8;
  k := 1;
  for i := 0 to CL.Count-1 do
  begin
    s := CL.Items[ i ];
    param := Trim( Parse( s, '=' ) );
    s := Trim( s );
    off := Str2Int( s );
    if off = j then
    begin
      Result := Result + '; p' + Int2Str( k ) + param + ': Integer';
      inc( j, 4 );
      inc( k );
    end;
  end;
  if Result = '' then Exit;
  Delete( Result, 1, 1 );
  Result := '(' + Result + ' )';
end;

function ParseLeft( var s: String; const Separators: String ): String;
var Pos : Integer;
begin
  Pos := IndexOfCharsMin( S, Separators );
  if Pos <= 0 then
     Pos := Length( S ) + 1;
  Result := S;
  S := Copy( Result, Pos, MaxInt );
  Result := Copy( Result, 1, Pos - 1 );
end;

function DataDir2Type( const s: String ): String;
begin
  if StrEq( s, 'DB' ) then Result := 'BYTE'
  else
  if StrEq( s, 'DW' ) then Result := 'WORD'
  else
  if StrEq( s, 'DD' ) then Result := 'Integer'
  else
  if StrEq( s, 'DQ' ) then Result := 'Int64'
  else
  asm
    int 3
  end;
end;

function CreateVariable( const s: String; var size_init: Integer ): String;
var nm, data_directive, cnt, dup: String;
    t: String;
    n: Integer;
begin
  size_init := 0;
  t := s;
  data_directive := SecondWord( t );
  nm := Parse( t, ' '#9 ); t := Trim( t );
  Parse( t, ' '#9 ); t := Trim( t );
  Result := 'var ' + ReplaceSymbols( nm ) + ': array[0..';
  cnt := Trim( Parse( t, ' '#9 ) ); t := Trim( t );
  n := 0;
  if cnt <> '' then
  begin
    if cnt[ Length( cnt ) ] in [ 'h', 'H' ] then
      n := Hex2Int( cnt )
    else
      n := Str2Int( cnt );
  end;
  dup := Trim( ParseLeft( t, ' '#9'(' ) ); t := Trim( t );
  if StrEq( dup, 'DUP' ) then
  begin
    Result := Result + Int2Str( n-1 ) + '] of ' + DataDir2Type( data_directive ) + ';';
    if StrIsStartingFrom( PChar( t ), '(?)' ) or
       StrIsStartingFrom( PChar( t ), '(0)' ) then
    else
      size_init := n;
  end
    else
  begin
    size_init := -1;
  end;
end;

function J_name( const s: String ): String;
begin
  Result := s;
  {if (Result <> '') and (Result[ 1 ] = '_') then
    //while (Result <> '') and (Result[ 1 ] = '_') do
      Delete( Result, 1, 1 )
  else}
    Result := 'j_' + Result;
end;

procedure ConvertFile( const fname, SecondUsesList: String );
var SL, UL, VL, IL, CL, PL, NL: PStrList;
    i, j: Integer;
    s, w, ww, f, t, fn, j_fn: String;
    in_proc, do_asm, forced_proc: Boolean;
    is_public: Boolean;
begin
  if Step = 1 then
    Write( 'Scanning ' + fname )
  else
    Write( 'Converting ' + fname );
  SL := NewStrList; //  ( )
  UL := NewStrList; // :  
  VL := NewStrList; // :   implementation (uses #2)
  IL := NewStrList; // :  implementation
  CL := NewStrList; // :      
  BL := NewStrList; // :     
  PL := NewStrList; //      
  EL := NewStrList; //  EXTRN- ( )
  NL := nil;        //      
  if (Step = 2) and (ML <> nil) then
  begin //     
    i := ML.IndexOf( ExtractFileNameWOext( fname ) );
    if i >= 0 then
    begin
      NL := Pointer( ML.Objects[ i ] ); //  
      NL := Pointer( NL.Tag ); //  
      NL := Pointer( NL.Tag ); //    
    end;
  end;
  TRY
    SL.LoadFromFile( fname );
    UL.Add( 'unit ' + ExtractFileNameWOext( fname ) + '; // automatically generated with Casm2Basm' );
    UL.Add( '' );
    UL.Add( 'interface' );
    UL.Add( '' );
    UL.Add( 'uses Windows, Messages;' );
    UL.Add( '' );

    IL.Add( '' );

    VL.Add( '' );
    if SecondUsesList <> '' then
    begin
      VL.Add( 'uses {$IFDEF ADDITIONAL_USES} ' +
              '{$I ' + ExtractFileNameWOext( fname ) + '_add2uses.inc} ' +
              '{$ENDIF} ' );
      VL.Add( '  ' + SecondUsesList );
    end
      else
    begin
      VL.Add( '{$IFDEF ADDITIONAL_USES} ' +
              '{$I ' + ExtractFileNameWOext( fname ) + '_add2uses.inc} ' +
              '{$ENDIF} ' );
    end;
    VL.Add( '' );
    s := ExtractFilePath( fname ) + ExtractFileNameWOext( fname ) + '_add2uses.inc';
    if not FileExists( s ) then
      CreateEmptyFile( s );

    i := 0;
    in_proc := FALSE;
    forced_proc := FALSE;
    is_data_segment := FALSE;
    while i <= SL.Count-1 do
    begin
      TRY
        s := SL.Items[ i ];
        if s = '' then continue;
        if s[ 1 ] = ';' then s := '// ' + s
        else
        begin
          do_asm := FALSE;
          w := FirstWord( s );
          if s[ 1 ] in [ ' ', #9 ] then
          begin //    
            if StrIn( w, [ 'TITLE', '.386P', 'ASSUME' ] ) then
              s := '// ' + s
            else
            begin
              ww := SecondWord( s );
              if not in_proc and StrEq( ww, 'ORG' )  then
                s := '// ' + s // ORG  
              else
                do_asm := TRUE;
            end;
          end
            else
          begin //    w <> ''
            if in_proc and forced_proc then
            begin
              IL.Add( 'end;' );
              IL.Add( '' );
              in_proc := FALSE;
              forced_proc := FALSE;
            end;

            if StrIn( w, [ 'include', '.model', 'INCLUDELIB', 'END',
                           'if', 'endif', 'else' ] ) then
              s := '// ' + s
            else
            if StrIn( w, [ 'EXTRN' ] ) then
            begin
              f := s;
              Parse( f, ' '#9 );
              f := ReplaceSymbols( Trim( Parse( f, ':' ) ) );
              if (f <> '') {and (f[ 1 ] = '_')} then
                EL.Add( f );
              s := '// ' + s;
            end
            else
            if StrIn( w, [ 'PUBLIC' ] ) then
            begin //   /( )?
              f := s;
              Parse( f, ' '#9 ); f := Trim( f );
              f := ReplaceSymbols( Trim( Parse( f, ' ;:'#9 ) ) );
              s := '// ' + s;
              //UL.Add( 'procedure ' + f + ';' );
              PL.Add( f );
            end
              else
            if StrIn( w, [ 'COMM' ] ) then
            begin
              f := CommArrayDeclaration( s );
              UL.Add( f );
              s := '// ' + s;
            end
              else
            begin
              ww := SecondWord( s );
              if StrIn( ww, [ 'GROUP', 'ENDP' ] ) then
                s := '// ' + s
              else
              if StrIn( ww, [ 'SEGMENT' ] ) then
              begin
                is_data_segment := StrIn( w, [ '_BSS' ] );
                s := '// ' + s;
              end
              else
              if StrIn( ww, [ 'ENDS' ] ) then
              begin //    (   -)
                s := '// ' + s;
                if in_proc then
                begin
                  in_proc := FALSE;
                  forced_proc := FALSE;
                  IL.Add( 'end;' );
                  IL.Add('');
                  CL.Clear;
                  BL.Clear;
                end;
              end
              else
              if StrIn( ww, [ 'PROC' ] ) then
              begin //    / 
                Assert( not in_proc, 'Leveled procedure -- not allowed' );
                in_proc := TRUE;
                IL.Add( '' );
                f := ReplaceSymbols( w );

                is_public := PL.IndexOf( f ) >= 0;
                fn := 'procedure ' + f + ';';
                if is_public then
                begin
                  j_fn := 'function ' + J_name( f ) +
                       ParamList( CL ) + ': Integer; cdecl;';
                  IL.Add( j_fn );
                  IL.Add( 'asm pop ebp' );
                  IL.Add( '    jmp ' + f );
                  IL.Add( 'end;' );
                end;
                IL.Add( fn );
                if is_public then
                begin
                  UL.Add( j_fn );
                  UL.Add( fn );
                end
                else
                  VL.Add( fn + ' forward;' );
                if CL.Count > 0 then
                begin
                  IL.Add( 'const' );
                  IL.AddStrings( CL );
                end;
                IL.Add( 'asm' );
                s := '// ' + s;
              end
                else
              if ww = '=' then
              begin //   
                Assert( not in_proc, 'Offset constant must be before PROC' );
                if s[ 1 ] = '$' then
                begin //     
                  BL.Add( w );
                  s := '_' + s;
                end;
                t := s;
                f := '  ' + ReplaceSymbols( Trim( Parse( t, ';' ) ) ) + ';';
                if Trim( t ) <> '' then
                  f := f + ' //' + t;
                CL.Add( f );
                s := '';
              end
                else
              if not in_proc and StrIn( ww, [ 'DB', 'DD', 'DW', 'DQ' ] ) then
              begin
                j := 1;
                if is_data_segment then
                begin
                  f := CreateVariable( s, j );
                  if j = 0 then s := '// ' + s;
                  UL.Add( f );
                  do_asm := FALSE;
                end;
                if j <> 0 then
                begin
                  in_proc := TRUE;
                  forced_proc := TRUE;
                  f := ReplaceSymbols( w );
                  if is_data_segment then
                  begin
                    f := 'init_' + f;
                    asm
                      int 3 //   ,    !
                    end;
                  end;
                  IL.Add( 'procedure ' + f + ';' );
                  if is_data_segment or (PL.IndexOf( f ) < 0) then
                    VL.Add( 'procedure ' + f + '; forward;' )
                  else
                    UL.Add( 'procedure ' + f + '; // initialized constant, not a procedure to call' );
                  IL.Add( 'asm' );
                  Parse( s, ' '#9 );
                  do_asm := TRUE;
                end;
              end
                else
              if not in_proc and StrIn( ww, [ 'ORG' ] ) then
              begin // ORG   -   , .
                s := '// ' + s;
              end
                else
              begin //    - -?
                do_asm := TRUE;
              end;
            end;
          end;
          if do_asm then
          begin
            ww := AsmOperator( s );
            if StrIn( ww, [ 'jmp', 'je', 'jz', 'jne', 'jnz', 'jl', 'jle', 'jnl',
                            'jb', 'jbe', 'jnb', 'jnbe', 'jg', 'jge', 'jng',
                            'ja', 'jae', 'jna', 'jnae', 'jp', 'jpo', 'jpe', 'jo',
                            'jno', 'jnp', 'js', 'jns',
                            'jc', 'jcxz', 'jecxz'  ] ) then
            begin
              s := RemoveShort( s );
            end
              else
            if StrIn( ww, [ 'ORG' ] ) then
            begin
              s := ReplaceOrg( s );
            end
              else
            if StrIn( ww, [ 'DQ', 'DD' ] ) then
            begin
              s := ReplaceDQR( s, ww );
            end
              else
            if StrIn( ww, [ 'test' ] ) then
            begin
              s := SwapTestOperands( s );
            end
              else
            if StrIn( ww, [ 'npad' ] ) then
            begin
              s := ReplaceNPAD( s );
            end;
            if pos( ';', s ) > 0 then //  
            begin
              t := s;
              f := PartBeforeSemicolon( t );
              f := RemoveFlat( f );
              if t <> '' then
                s := ReplaceSymbols( f ) + ' //' + t
              else
                s := ReplaceSymbols( f );
            end
              else
            begin
              s := RemoveFlat( s );
              s := ReplaceSymbols( s );
            end;
          end;
        end;
        if s <> '' then
          IL.Add( s );
      FINALLY
        inc( i );
      END;
    end;
    Assert( not in_proc, 'End of PROC not found' );

    IL.Add( '' );
    IL.Add( 'end.' );

    {$IFDEF GENERATE_TEST}
    if Step = 2 then
    if EL.Count > 0 then
    begin
      VL.Add( '' );
      VL.Add( '{$IFDEF TESTING}' );
      for i := 0 to EL.Count-1 do
      begin
        if (NL <> nil) and (NL.IndexOf( EL.Items[ i ] ) >= 0) or (ML = nil) then
          VL.Add( 'procedure ' + EL.Items[ i ] + '; begin end;' );
      end;
      VL.Add( '{$ENDIF TESTING}' );
    end;
    {$ENDIF}

    if Step = 1 then
    begin //    
      if ML = nil then
        ML := NewStrListEx;
      ML.AddObject( ExtractFileNameWOext( fname ), DWORD( PL ) );
      PL.Tag := DWORD( EL );
      PL := nil;
      EL := nil;
    end;

    if Step = 2 then
    begin //  pas-
      UL.Add( '' );
      UL.Add( 'implementation' );
      UL.AddStrings( VL );
      UL.AddStrings( IL );
      UL.SaveToFile( ReplaceExt( fname, '.pas' ) );
    end;

    Writeln( ' - OK.' );
  FINALLY
    SL.Free;
    UL.Free;
    VL.Free;
    IL.Free;
    CL.Free;
    EL.Free;
    PL.Free;
    BL.Free;
  END;
  asm
    jmp @@1
    DB 0ebh, 03h, 8dh, 49h, 00h
  @@1:
  end;
end;

function UL2uses2( UL: PStrList ): String;
var FL: PStrList;
    s: String;
    i: Integer;
begin
  Result := '';
  if UL.Count = 0 then Exit;
  FL := NewStrList;
  TRY
    s := '';
    for i := 0 to UL.Count-1 do
    begin
      s := s + UL.Items[ i ] + ', ';
      if Length( s ) > 64 then
      begin
        FL.Add( s );
        s := '  ';
      end;
    end;
    if s <> '' then FL.Add( s );
    Result := Trim( FL.Text );
    Delete( Result, Length( Result ), 1 ); //   
    Result := Result + ';';
  FINALLY
    FL.Free;
  END;
end;

//  uses #2   .
function BuildUses2( const module: String ): String;
var m_EL, m_IL: PStrList;
    i, j, k: Integer;
    UL, NL: PStrList;
    imported_name: String;
    found: Boolean;
begin
  Result := '';
  i := ML.IndexOf( module );
  if i < 0 then
  begin
    asm
      int 3 //    
    end;
    Exit;
  end;
  m_EL := Pointer( ML.Objects[ i ] );
  {$IFDEF SAVE_LISTS}
  m_EL.SaveToFile( GetStartDir + module + '.exported.txt' );
  {$ENDIF}

  m_IL := Pointer( m_EL.Tag ); //       
  {$IFDEF SAVE_LISTS}
  m_IL.SaveToFile( GetStartDir + module + '.imported.txt' );
  {$ENDIF}

  UL := NewStrList; //   ,     uses #2
  NL := NewStrList; //       module (.. 
                    //   module,        )
  TRY
    for k := 0 to m_IL.Count-1 do
    begin
      imported_name := m_IL.Items[ k ];
      found := FALSE;
      for j := 0 to ML.Count-1 do
        if i <> j then //     
        begin
          m_EL := Pointer( ML.Objects[ j ] );
          if m_EL.IndexOf( imported_name ) >= 0 then
          begin //  ,     
            found := TRUE;
            if UL.IndexOf( ML.Items[ j ] ) < 0 then
              UL.Add( ML.Items[ j ] );
            break;
          end;
        end;
      if not found then
        NL.Add( imported_name );
    end;
    {$IFDEF SAVE_LISTS}
    NL.SaveToFile( GetStartDir + module + '.notresolved.txt' );
    {$ENDIF}
    if NL.Count > 0 then
    begin
      m_IL.Tag := DWORD( NL );
      NL := nil;
    end;
    //    uses
    Result := UL2uses2( UL );
  FINALLY
    UL.Free;
    NL.Free;
  END;
end;

procedure ClearModiles; //    
var i: Integer;
    m_EL: PStrList; //    (PUBLIC)
    m_IL: PStrList; //    (EXTRN)
    m_NL: PStrList; //      (EXTRN)
begin
  if ML = nil then Exit;
  for i := 0 to ML.Count-1 do
  begin
    m_EL := Pointer( ML.Objects[ i ] );
    m_IL := Pointer( m_EL.Tag );
    m_NL := Pointer( m_IL.Tag );
    m_EL.Free;
    m_IL.Free;
    m_NL.Free;
  end;
  ML.Free;
end;

initialization

finalization
  ClearModiles;

end.
